﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

#if !NO_WEBAPI
using System.Web.Http.Dependencies;
#endif

namespace Simple.Resolver
{
    /// <summary>
    /// Simple implementation of the IDependencyResolver interface. (Web API)
    /// All instances are created "per scope", no builtin support for singletons and transient instances. (easily solved by always returning same or new instance)
    /// Resolve methods does not mutate the container and are threadsafe.
    /// Register methods are not thread safe.
    /// </summary>
    public class DependencyResolver : IDependencyResolver
    {
        const string AppSettingKeyFormat = "{0}.{1}";


        IDictionary<Type, ResolverDelegate> resolvers;
        IDictionary<Type, ParseAppSettingDelegate> appSettingParsers;
        IDictionary<Type, object> instances = new Dictionary<Type, object>();


        public DependencyResolver()
        {
            resolvers = new Dictionary<Type, ResolverDelegate>();
            appSettingParsers = new Dictionary<Type, ParseAppSettingDelegate>();

            RegisterReadFromAppSettingType(value => value);
        }

        DependencyResolver(IDictionary<Type, ResolverDelegate> resolvers, IDictionary<Type, ParseAppSettingDelegate> appSettingParsers)
        {
            this.resolvers = resolvers;
            this.appSettingParsers = appSettingParsers;
        }

        public DependencyResolver Register<T>(Func<DependencyResolver, T> resolver)
        {
            Argument.NotNull(resolver, "resolver");

            resolvers[typeof(T)] = r => resolver(r);
            return this;
        }

        public DependencyResolver Register<T>()
        {
            return Register<T, T>();
        }

        public DependencyResolver Register<T, TImplementation>()
            where TImplementation : T
        {
            if (typeof(TImplementation).IsAbstract)
            {
                if (typeof(TImplementation) == typeof(T))
                    throw new ArgumentException("Type " + typeof(TImplementation) + " is abstract and can not be instantiated");

                return Register<T>(r => r.Resolve<TImplementation>());
            }

            resolvers[typeof(T)] = CreateDefaultResolverDelegate(typeof(TImplementation));
            return this;
        }

        public DependencyResolver Register(Assembly assembly, Type ofType = null)
        {
            Argument.NotNull(assembly, "assembly");

            var typesToRegister =
                from t in assembly.GetExportedTypes()
                where t.IsClass && !t.IsAbstract && GetSuitableConstructor(t) != null && (ofType == null || ofType.IsAssignableFrom(t))
                select t;

            foreach (var type in typesToRegister)
            {
                resolvers[type] = CreateDefaultResolverDelegate(type);
            }

            return this;
        }

        public DependencyResolver RegisterReadFromAppSettingType<T>(Func<string, T> parser)
        {
            Argument.NotNull(parser, "parser");

            appSettingParsers[typeof(T)] = value => parser(value);
            return this;
        }


        public T Resolve<T>()
        {
            var instance = TryResolve(typeof(T));

            if (instance == null)
                throw new InvalidOperationException("Can not resolve type " + typeof(T));

            return (T)instance;
        }

        public T GetAppSettingValue<T>(string key)
        {
            var value = ConfigurationManager.AppSettings[key];

            if (value == null)
                throw new ConfigurationErrorsException(string.Format("App setting value for key '{0}' is missing", key));

            ParseAppSettingDelegate parser;
            if (!appSettingParsers.TryGetValue(typeof(T), out parser))
                throw new InvalidOperationException();

            try
            {
                return (T)parser(value);
            }
            catch (Exception ex)
            {
                throw new ConfigurationErrorsException(string.Format("App setting value for key '{0}' is invalid: '{1}", key, value), ex);
            }
        }

        public T GetAppSettingValueOrResolve<T>(string key)
        {
            return
                appSettingParsers.ContainsKey(typeof(T))
                ? GetAppSettingValue<T>(key)
                : Resolve<T>();
        }

        object TryResolve(Type type)
        {
            if (type == typeof(IDependencyScope) ||
                type == typeof(IDependencyResolver) ||
                type == typeof(DependencyResolver))
                return this;

            lock (instances)
            {
                object instance;

                if (instances.TryGetValue(type, out instance))
                    return instance;

                ResolverDelegate resolver;

                if (!resolvers.TryGetValue(type, out resolver))
                    return null;

                instance = resolver(this);
                instances.Add(type, instance);

                return instance;
            }
        }

        ResolverDelegate CreateDefaultResolverDelegate(Type type)
        {
            var constructor = GetSuitableConstructor(type);

            if (constructor == null)
                throw new ArgumentException("Type " + type + " has no suitable public constructor and can not be instantiated");

            var resolverParameter =
                Expression.Parameter(typeof(DependencyResolver));

            var constructorParameters =
                constructor.
                    GetParameters().
                    Select(
                        p =>
                        Expression.Call(
                            resolverParameter,
                            "GetAppSettingValueOrResolve",
                            new[] { p.ParameterType },
                            Expression.Constant(string.Format(AppSettingKeyFormat, type.Name, p.Name))));

            return Expression.Lambda<ResolverDelegate>(Expression.New(constructor, constructorParameters), resolverParameter).Compile();
        }

        static ConstructorInfo GetSuitableConstructor(Type type)
        {
            return
                type.GetConstructors().
                     Where(c => c.IsPublic && !c.GetParameters().Any(p => p.ParameterType == type)).
                     OrderByDescending(c => c.GetParameters().Length).
                     FirstOrDefault();
        }



        object IDependencyScope.GetService(Type serviceType)
        {
            Argument.NotNull(serviceType, "serviceType");

            return TryResolve(serviceType);
        }

        IEnumerable<object> IDependencyScope.GetServices(Type serviceType)
        {
            Argument.NotNull(serviceType, "serviceType");

            var instance = TryResolve(serviceType);

            return
                instance != null
                ? new[] { instance }
                : Enumerable.Empty<object>();
        }

        IDependencyScope IDependencyResolver.BeginScope()
        {
            return new DependencyResolver(resolvers, appSettingParsers);
        }

        void IDisposable.Dispose()
        {
            lock (instances)
            {
                foreach (var disposable in instances.Values.OfType<IDisposable>())
                {
                    disposable.Dispose();
                }
            }
        }

        delegate object ResolverDelegate(DependencyResolver resolver);

        delegate object ParseAppSettingDelegate(string value);

        static class Argument
        {
            public static void NotNull<T>(T value, string paramName) where T : class
            {
                if (value == null)
                    throw new ArgumentNullException(paramName);
            }
        }
    }
}
